""" tests types """

import pytest
import pydantic.error_wrappers

from kanidm.types import AuthInitResponse, KanidmClientConfig, RadiusGroup, RadiusClient


def test_auth_init_response() -> None:
    """tests AuthInitResponse"""
    testobj = {
        "sessionid": "crabzrool",
        "state": {
            "choose": ["passwordmfa"],
        },
    }

    testval = AuthInitResponse.parse_obj(testobj)
    assert testval.sessionid == "crabzrool"


def test_radiusgroup_vlan_negative() -> None:
    """tests RadiusGroup's vlan validator"""
    with pytest.raises(pydantic.error_wrappers.ValidationError):
        RadiusGroup(vlan=-1)


def test_radiusgroup_vlan_zero() -> None:
    """tests RadiusGroup's vlan validator"""
    with pytest.raises(pydantic.error_wrappers.ValidationError):
        RadiusGroup(vlan=0)


def test_radiusgroup_vlan_4096() -> None:
    """tests RadiusGroup's vlan validator"""
    assert RadiusGroup(vlan=4096, name="crabzrool")


def test_radiusgroup_vlan_no_name() -> None:
    """tests RadiusGroup's vlan validator"""
    with pytest.raises(
        pydantic.error_wrappers.ValidationError, match="name\n.*field required"
    ):
        RadiusGroup(
            vlan=4096,
        )


def test_kanidmconfig_parse_toml() -> None:
    """tests KanidmClientConfig.parse_toml()"""

    config = KanidmClientConfig()
    config.parse_toml("uri = 'https://crabzrool.example.com'")

def test_radius_client_bad_hostname() -> None:
    """tests with a bad hostname"""
    with pytest.raises(pydantic.error_wrappers.ValidationError):
        RadiusClient(name="test", ipaddr="thiscannotpossiblywork.kanidm.example.com",secret="nothing")

    assert RadiusClient(name="test", ipaddr="kanidm.com",secret="nothing")



""" tests ssl validation and CA setting etc """

from pathlib import Path

import aiohttp
import aiohttp.client_exceptions

import pytest


from kanidm import KanidmClient


@pytest.mark.asyncio
async def test_ssl_valid() -> None:
    """tests a valid connection"""

    url = "https://badssl.com"

    async with aiohttp.ClientSession() as session:
        client = KanidmClient(
            uri=url,
            session=session,
        )

        result = await client.call_get("/")
        assert result.content
        print(f"{result.status_code=}")


@pytest.mark.asyncio
async def test_ssl_self_signed() -> None:
    """tests with a self-signed cert"""

    url = "https://self-signed.badssl.com"

    async with aiohttp.ClientSession() as session:
        print("testing self signed cert with defaults and expecting an error")
        client = KanidmClient(
            uri=url,
            session=session,
        )
        with pytest.raises(aiohttp.client_exceptions.ClientConnectorCertificateError):
            await client.call_get("/")


@pytest.mark.asyncio
async def test_ssl_self_signed_with_verify() -> None:
    """tests with a self-signed cert"""

    async with aiohttp.ClientSession() as session:
        client = KanidmClient(
            uri="https://self-signed.badssl.com",
            session=session,
            verify_certificate=False,
        )
        result = await client.call_get("/")
        assert result.content


@pytest.mark.asyncio
async def test_ssl_self_signed_no_verify_certificate() -> None:
    """tests with a self-signed cert"""

    async with aiohttp.ClientSession() as session:
        client = KanidmClient(
            uri="https://self-signed.badssl.com",
            session=session,
            verify_certificate=False,
        )
        result = await client.call_get("/")
        assert result.content


@pytest.mark.asyncio
async def test_ssl_wrong_hostname_throws_error() -> None:
    """tests with validate hostnames and wrong hostname in the cert"""

    async with aiohttp.ClientSession() as session:
        client = KanidmClient(
            uri="https://wrong.host.badssl.com/", session=session, verify_hostnames=True
        )
        with pytest.raises(
            aiohttp.client_exceptions.ClientConnectorCertificateError,
            match="Cannot connect to host wrong.host.badssl.com:443",
        ):
            result = await client.call_get("/")
            assert result.content


@pytest.mark.asyncio
async def test_ssl_wrong_hostname_dont_verify_hostnames() -> None:
    """tests with validate hostnames and wrong hostname in the cert"""

    async with aiohttp.ClientSession() as session:
        client = KanidmClient(
            uri="https://wrong.host.badssl.com/",
            session=session,
            verify_hostnames=False,
        )
        result = await client.call_get("/")
        assert result.content


@pytest.mark.asyncio
async def test_ssl_wrong_hostname_verify_certificate() -> None:
    """tests with validate hostnames and wrong hostname in the cert"""

    async with aiohttp.ClientSession() as session:
        client = KanidmClient(
            uri="https://wrong.host.badssl.com/",
            session=session,
            verify_hostnames=False,
            verify_certificate=False,
        )
        result = await client.call_get("/")
        assert result.content


@pytest.mark.asyncio
async def test_ssl_revoked() -> None:
    """tests with a revoked certificate, it'll pass but one day this should be a thing"""

    async with aiohttp.ClientSession() as session:
        client = KanidmClient(
            uri="https://revoked.badssl.com/",
            session=session,
        )
        result = await client.call_get("/")
        assert result.content


@pytest.mark.asyncio
async def test_ssl_expired() -> None:
    """tests with an expired certificate"""

    async with aiohttp.ClientSession() as session:
        client = KanidmClient(
            uri="https://expired.badssl.com/",
            session=session,
        )
        with pytest.raises(
            aiohttp.client_exceptions.ClientConnectorCertificateError,
            match="certificate verify failed: certificate has expired",
        ):
            result = await client.call_get("/")
            assert result.content


@pytest.mark.asyncio
async def test_ssl_expired_ignore() -> None:
    """tests with an expired certificate"""

    async with aiohttp.ClientSession() as session:
        client = KanidmClient(
            uri="https://expired.badssl.com/",
            session=session,
            verify_certificate=False,
        )
        result = await client.call_get("/")
        assert result.content


@pytest.mark.asyncio
async def test_ssl_untrusted_root_throws() -> None:
    """tests with an untrusted root, which should throw an error"""

    async with aiohttp.ClientSession() as session:
        client = KanidmClient(
            uri="https://untrusted-root.badssl.com/",
            session=session,
        )
        with pytest.raises(
            aiohttp.client_exceptions.ClientConnectorCertificateError,
            match="certificate verify failed: self signed certificate in certificate chain",
        ):
            result = await client.call_get("/")
            assert result.content


@pytest.mark.asyncio
async def test_ssl_untrusted_root_configured() -> None:
    """tests with an untrusted root, which should throw an error"""

    testcert = Path("./tests/badssl_trusted_ca.pem").resolve()

    if not testcert.exists():
        pytest.skip(f"The trusted cert is missing from {testcert}")

    async with aiohttp.ClientSession() as session:
        client = KanidmClient(
            uri="https://untrusted-root.badssl.com/",
            session=session,
            ca_path=testcert.resolve().as_posix(),
        )
        with pytest.raises(
            aiohttp.client_exceptions.ClientConnectorCertificateError,
            match="certificate verify failed: self signed certificate in certificate chain",
        ):
            result = await client.call_get("/")
            assert result.content



""" reusable widgets for testing """

from pathlib import Path
from typing import Any

import pytest
from kanidm import KanidmClient


@pytest.fixture(scope="function")
async def client() -> KanidmClient:
    """sets up a client with a basic thing"""
    try:
        return KanidmClient(uri="https://idm.example.com")
    except FileNotFoundError:
        raise pytest.skip("Couldn't find config file...")


@pytest.fixture(scope="function")
async def client_configfile() -> KanidmClient:
    """sets up a client from a config file"""
    try:
        return KanidmClient(config_file=Path("~/.config/kanidm"))
    except FileNotFoundError:
        raise pytest.skip("Couldn't find config file...")


class MockResponse:
    """mock the things"""

    def __init__(self, text: str, status: int) -> None:
        self._text = text
        self.status = status

    async def text(self) -> str:
        """mock the things"""
        return self._text

    # pylint: disable=invalid-name
    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        """mock the things"""

    async def __aenter__(self) -> Any:
        """mock the things"""
        return self



""" testing auth things """

import logging
import os

import aiohttp
import pytest
from pytest_mock import MockerFixture

# pylint: disable=unused-import
from testutils import client, client_configfile, MockResponse

from kanidm import KanidmClient
from kanidm.exceptions import AuthCredFailed, AuthInitFailed
from kanidm.types import AuthBeginResponse


logging.basicConfig(level=logging.DEBUG)


@pytest.mark.asyncio
async def test_auth_init(client_configfile: KanidmClient) -> None:
    """tests the auth init step"""
    print("Starting client...")
    print(f"Doing auth_init for {client_configfile.config.username}")

    if client_configfile.config.username is None:
        raise ValueError("This path shouldn't be possible in the test!")
    async with aiohttp.ClientSession() as session:
        client_configfile.session = session
        result = await client_configfile.auth_init(client_configfile.config.username)
    print(f"{result=}")
    print(result.dict())
    assert result.sessionid


@pytest.mark.asyncio
async def test_auth_begin(client_configfile: KanidmClient) -> None:
    """tests the auth begin step"""
    print(f"Doing auth_init for {client_configfile.config.username}")

    async with aiohttp.ClientSession() as session:
        client_configfile.session = session
        if client_configfile.config.username is None:
            raise ValueError("This path shouldn't be possible in the test!")
        result = await client_configfile.auth_init(client_configfile.config.username)
        print(f"{result=}")
        print("Result dict:")
        print(result.dict())
        assert result.sessionid

        print(f"Doing auth_begin for {client_configfile.config.username}")
        begin_result = await client_configfile.auth_begin(
            # username=client.username,
            method="password",
        )
        print(f"{begin_result=}")
        print(begin_result.data)
        retval = begin_result.data

        if retval is None:
            raise pytest.fail("Failed to do begin_result")

        retval["response"] = begin_result

        assert AuthBeginResponse.parse_obj(retval)


@pytest.mark.asyncio
async def test_authenticate_flow(client_configfile: KanidmClient) -> None:
    """tests the authenticate() flow"""
    async with aiohttp.ClientSession() as session:
        print(f"Doing client.authenticate for {client_configfile.config.username}")
        client_configfile.session = session
        result = await client_configfile.authenticate_password()
    print(result)


@pytest.mark.asyncio
async def test_authenticate_flow_fail(client_configfile: KanidmClient) -> None:
    """tests the authenticate() flow with a valid (hopefully) usernamd and invalid password"""
    if not bool(os.getenv("RUN_SCARY_TESTS", None)):
        pytest.skip(reason="Skipping because env var RUN_SCARY_TESTS isn't set")
    print("Starting client...")
    if (
        client_configfile.config.uri is None
        or client_configfile.config.username is None
        or client_configfile.config.password is None
    ):
        pytest.skip("Please ensure you have a username, password and uri in the config")
    print(f"Doing client.authenticate for {client_configfile.config.username}")

    async with aiohttp.ClientSession() as session:
        client_configfile.session = session
        with pytest.raises((AuthCredFailed, AuthInitFailed)):
            result = await client_configfile.authenticate_password(
                username=client_configfile.config.username,
                password="cheese",
            )
            print(result)


# TODO: mock a call to auth_init when a 200 response is not returned, raises AuthInitFailed
# TODO: mock a call to auth_init when "x-kanidm-auth-session-id" not in response.headers, raises ValueError


# TODO: mock a call to auth_begin when a 200 response is not returned, raises AuthBeginFailed
# TODO: mock a call to auth_step_password when a 200 response is not returned, raises AuthCredFailed


@pytest.mark.asyncio
async def test_authenticate_inputs_validation(
    client: KanidmClient, mocker: MockerFixture
) -> None:
    """tests if you pass username but not password and password but not username"""

    resp = MockResponse("crabs are cool", 200)

    mocker.patch("aiohttp.ClientSession.post", return_value=resp)

    async with aiohttp.ClientSession() as session:
        client.session = session
        with pytest.raises(ValueError):
            await client.authenticate_password(username="cheese")
        with pytest.raises(ValueError):
            await client.authenticate_password(password="cheese")
        client.config.password = None
        client.config.username = "crabby"
        with pytest.raises(ValueError):
            await client.authenticate_password()
        client.config.password = "cR4bzR0ol"
        client.config.username = None
        with pytest.raises(ValueError):
            await client.authenticate_password()

        client.config.username = None
        client.config.password = None
        with pytest.raises(ValueError):
            await client.authenticate_password()


@pytest.mark.asyncio
async def test_auth_step_password(client: KanidmClient) -> None:
    """tests things"""

    with pytest.raises(ValueError):
        async with aiohttp.ClientSession() as session:
            client.session = session
            await client.auth_step_password()



""" utility functions """

from pathlib import Path
from typing import Any, Dict, Union

import toml


def load_config(filename: Union[str, Path] = "/etc/kanidm/config") -> Dict[str, Any]:
    """loads the configuration file"""
    if isinstance(filename, Path):
        config_filepath = filename
    else:
        config_filepath = Path(filename).expanduser().resolve()

    if not config_filepath.exists():
        raise FileNotFoundError(
            f"Failed to find configuration file ({config_filepath}), quitting!",
        )
    config_data: Dict[str, Any] = toml.load(config_filepath.open(encoding="utf-8"))
    return config_data



""" testing session header function """

import pytest

import aiohttp.client_exceptions
from testutils import client

from kanidm import KanidmClient


def test_session_header(client: KanidmClient) -> None:
    """tests the session_header function"""

    with pytest.raises(ValueError):
        client.session_header()

    assert client.session_header("testval") == {
        "X-KANIDM-AUTH-SESSION-ID": "testval",
    }


@pytest.mark.asyncio
async def test_session_creator(client: KanidmClient) -> None:
    """tests the session_header function"""

    client.session = None
    client.config.uri = "🦀"
    with pytest.raises(aiohttp.client_exceptions.InvalidURL):
        await client._call(method="GET", path="/")  # pylint: disable=protected-access

    #  pytest.raises(ValueError):
    #     client.session_header()

    # assert client.session_header("testval") == {
    #     "X-KANIDM-AUTH-SESSION-ID": "testval",
    # }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kanidm', 'tests']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0', 'pydantic>=1.9.2,<2.0.0', 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'kanidm',
    'version': '0.0.2',
    'description': 'Kanidm client library',
    'long_description': "# kanidm\n\nA Python module for interacting with Kanidm.\n\nCurrently in very very very early beta, please [log an issue](https://github.com/kanidm/kanidm/issues/new/choose) for feature requests and bugs.\n\n## Installation\n\n```shell\npython -m pip install kanidm\n```\n\n## Documentation\n\nDocumentation can be generated by [cloning the repository](https://github.com/kanidm/kanidm) and running `make docs/pykanidm/build`. The documentation will appear in `./pykanidm/site`. You'll need make and the [poetry](https://pypi.org/project/poetry/) package installed.\n\n\n## Changelog\n\n| Version | Date | Notes |\n| ---     | ---  | ---   |\n| 0.0.1   | 2022-08-16 | Initial release |\n| 0.0.2   | 2022-08-16 | Updated license, including test code in package |\n",
    'author': 'James Hodgkinson',
    'author_email': 'james@terminaloutcomes.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://kanidm.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



""" testing get_radius_token """


import json
import logging
import os

# from typing import Any

import aiohttp
import pytest

# from pytest_mock import MockerFixture

# pylint: disable=unused-import
from testutils import client, client_configfile

from kanidm import KanidmClient

# from kanidm.exceptions import AuthCredFailed, AuthInitFailed
# from kanidm.types import AuthBeginResponse

logging.basicConfig(level=logging.DEBUG)


@pytest.mark.asyncio
async def test_radius_call(client_configfile: KanidmClient) -> None:
    """tests the radius call step"""
    print(f"Doing auth_init for {client_configfile.config.username}")

    if "RADIUS_USER" not in os.environ:
        pytest.skip(
            "Skipping this test - set RADIUS_USER environment variable to a valid RADIUS user."
        )

    radius_user = os.environ["RADIUS_USER"]

    if client_configfile.config.username is None:
        raise ValueError("This path shouldn't be possible in the test!")
    async with aiohttp.ClientSession() as session:
        client_configfile.session = session
        radius_session = await client_configfile.authenticate_password()

        result = await client_configfile.get_radius_token(
            radius_user, radius_session_id=radius_session.sessionid
        )

    print(f"{result=}")
    print(json.dumps(result.dict(), indent=4, default=str))



""" tests the config file things """

import logging
from pathlib import Path
import sys

import aiohttp
import pydantic
import pytest

from kanidm import KanidmClient
from kanidm.types import KanidmClientConfig
from kanidm.utils import load_config

logging.basicConfig(level=logging.DEBUG)

EXAMPLE_CONFIG_FILE = "../examples/config"


@pytest.fixture(scope="function")
async def client() -> KanidmClient:
    """sets up a client with a basic thing"""
    async with aiohttp.ClientSession() as session:
        return KanidmClient(
            uri="https://idm.example.com",
            session=session,
        )


def test_load_config_file() -> None:
    """tests that the file loads"""
    if not Path(EXAMPLE_CONFIG_FILE).expanduser().resolve().exists():
        print("Can't find client config file", file=sys.stderr)
        pytest.skip()
    print("Loading config file")
    config = load_config(EXAMPLE_CONFIG_FILE)
    assert config.get("uri") == "https://idm.example.com"

    print(f"{config.get('uri')=}")
    print(config)


def test_load_missing_config_file() -> None:
    """tests that an error is raised"""

    with pytest.raises(
        FileNotFoundError,
        match=EXAMPLE_CONFIG_FILE + "cheese",
    ):
        load_config(EXAMPLE_CONFIG_FILE + "cheese")


def test_parse_config_validationerror(client: KanidmClient) -> None:
    """tests parse_config with a faulty input"""
    testdict = {"verify_certificate": "that was weird."}
    with pytest.raises(ValueError):
        client.parse_config_data(config_data=testdict)


@pytest.mark.asyncio
async def test_parse_config_data(client: KanidmClient) -> None:
    """tests parse_config witha  valid input"""

    async with aiohttp.ClientSession() as session:
        client.session = session
        testdict = {
            "uri": "https://example.com",
            "username": "testuser",
            "password": "CraBzR0oL",
        }
        client.parse_config_data(config_data=testdict)


@pytest.mark.asyncio
async def test_init_with_uri() -> None:
    """tests the class"""

    async with aiohttp.ClientSession() as session:
        testclient = KanidmClient(
            uri="https://example.com",
            session=session,
        )
    assert testclient.config.uri == "https://example.com/"


@pytest.mark.asyncio
async def test_init_with_session() -> None:
    """tests the class"""
    async with aiohttp.ClientSession() as session:
        testclient = KanidmClient(
            uri="https://google.com",
            session=session,
        )
        assert testclient.session is session


def test_config_invalid_uri() -> None:
    """tests passing an invalid uri to the config parser"""

    test_input = {
        "uri": "asdfsadfasd",
    }
    with pytest.raises(pydantic.ValidationError):
        KanidmClientConfig.parse_obj(test_input)


def test_config_none_uri() -> None:
    """tests passing an invalid uri to the config parser"""

    with pytest.raises(ValueError, match="Please intitialize this with a server URI"):
        KanidmClient(uri=None)


def test_config_loader_str() -> None:
    """tests passing an invalid uri to the config parser"""

    with pytest.raises(FileNotFoundError):
        KanidmClient(config_file="hello world")


def test_config_init() -> None:
    """tests passing config object"""

    config = KanidmClientConfig(uri="https://idp.crabzrool.test")
    assert KanidmClient(config=config)
